#
# Copyright (c) 2019, salesforce.com, inc.
# All rights reserved.
# SPDX-License-Identifier: BSD-3-Clause
# For full license text, see the LICENSE file in the repo root or https://opensource.org/licenses/BSD-3-Clause
#

from django_declarative_apis.machinery.filtering import ALWAYS, NEVER, expandable

from . import models

DEFAULT_FILTERS = {
    str: ALWAYS,
    int: ALWAYS,
    dict: ALWAYS,
    models.TestModel: {
        'pk': ALWAYS,
        'int_field': ALWAYS,
        'expandable_dict': expandable(),
        'expandable_string': expandable()
    },
    models.ChildModel: {
        'pk': ALWAYS,
        'test': expandable(model_class=models.TestModel),
        'name': ALWAYS,
        'parent': expandable(model_class=models.ParentModel)
    },
    models.ParentModel: {
        'nonstandard_id': ALWAYS,
        'name': ALWAYS,
        'favorite': expandable(model_class=models.ChildModel, display_key='name'),
        'children': expandable(model_class=models.ChildModel, display_key='name'),
    },
    models.RootNode: {
        'pk': ALWAYS,
        'parent_field': expandable(model_class=models.ParentModel),
        'parents': expandable(model_class=models.ParentModel)
    }
}

DEFAULT_FILTERS_NO_EXPANDABLE = {
    str: ALWAYS,
    int: ALWAYS,
    dict: ALWAYS,
    models.TestModel: {
        'pk': ALWAYS,
        'int_field': ALWAYS,
        'expandable_dict': ALWAYS,
        'expandable_string': ALWAYS
    },

}



#
# Copyright (c) 2019, salesforce.com, inc.
# All rights reserved.
# SPDX-License-Identifier: BSD-3-Clause
# For full license text, see the LICENSE file in the repo root or https://opensource.org/licenses/BSD-3-Clause
#

import unittest

import django.test
import mock

from django_declarative_apis import models
from django_declarative_apis.authentication.oauthlib import request_validator

from tests import testutils


class DjangoRequestValidatorTestCase(django.test.TestCase):
    def setUp(self):
        self.consumer = models.OauthConsumer.objects.create(name='test')
        self.consumer.save()

    def test_get_client_secret_with_rsa_public_key_pem(self):
        self.consumer.rsa_public_key_pem = 'something non-null'

        request = django.test.RequestFactory().get('/')

        validator = request_validator.DjangoRequestValidator(request)
        validator.consumer = self.consumer
        self.assertIsNone(validator.get_client_secret(self.consumer.key, request))

    @mock.patch('django_declarative_apis.authentication.oauthlib.request_validator.logging')
    def test_get_rsa_key(self, mock_log):
        request = django.test.RequestFactory().get('/')

        validator = request_validator.DjangoRequestValidator(request)

        self.assertEqual(validator.get_rsa_key(self.consumer.key, request), '')
        mock_log.error.assert_called_with('This should never happen, since consumer is already validated')

        validator.consumer = self.consumer
        self.consumer.rsa_public_key_pem = 'something non-null'

        self.assertEqual(
            validator.get_rsa_key(self.consumer.key, request),
            self.consumer.rsa_public_key_pem
        )



#
# Copyright (c) 2019, salesforce.com, inc.
# All rights reserved.
# SPDX-License-Identifier: BSD-3-Clause
# For full license text, see the LICENSE file in the repo root or https://opensource.org/licenses/BSD-3-Clause
#

import base64
import logging
from oauthlib.oauth1.rfc5849 import SIGNATURE_RSA
from oauthlib.oauth1.rfc5849 import errors, signature
from oauthlib.oauth1 import SignatureOnlyEndpoint
from cryptography.hazmat.primitives.serialization import load_der_public_key, PublicFormat, Encoding
from cryptography.hazmat.backends import default_backend

from django_declarative_apis.resources.utils import preprocess_rsa_key

log = logging.getLogger(__name__)

class TweakedSignatureOnlyEndpoint(SignatureOnlyEndpoint):
    """An endpoint only responsible for verifying an oauthlib signature.
    This class modified oauthlib.oauth1.SignatureOnlyEndpoint so that
    the validate_request() method will support returning an error
    message to support our API OAuth error messages

    Altered lines are marked with # TOOPHER
    """


    def __init__(self, *args, **kwargs):
        super(TweakedSignatureOnlyEndpoint, self).__init__(*args, **kwargs)
        self.validation_error_message = ''

    def validate_request(self, uri, http_method='GET',
                         body=None, headers=None):
        """Validate a signed OAuth request.

        :param uri: The full URI of the token request.
        :param http_method: A valid HTTP verb, i.e. GET, POST, PUT, HEAD, etc.
        :param body: The request body as a string.
        :param headers: The request headers as a dict.
        :returns: A tuple of 2 elements.
                  1. True if valid, False otherwise.
                  2. An oauthlib.common.Request object.
        """
        try:
            request = self._create_request(uri, http_method, body, headers)
        except errors.OAuth1Error as e:
            return False, None

        try:
            self._check_transport_security(request)
            self._check_mandatory_parameters(request)
        except errors.OAuth1Error as e:
            self.validation_error_message = e.description  # TOOPHER
            return False, request

        if not self.request_validator.validate_timestamp_and_nonce(
                request.client_key, request.timestamp, request.nonce, request):
            return False, request

        # The server SHOULD return a 401 (Unauthorized) status code when
        # receiving a request with invalid client credentials.
        # Note: This is postponed in order to avoid timing attacks, instead
        # a dummy client is assigned and used to maintain near constant
        # time request verification.
        #
        # Note that early exit would enable client enumeration
        valid_client = self.request_validator.validate_client_key(
            request.client_key, request)
        if not valid_client:
            request.client_key = self.request_validator.dummy_client

        valid_signature = self._check_signature(request)

        # We delay checking validity until the very end, using dummy values for
        # calculations and fetching secrets/keys to ensure the flow of every
        # request remains almost identical regardless of whether valid values
        # have been supplied. This ensures near constant time execution and
        # prevents malicious users from guessing sensitive information
        v = all((valid_client, valid_signature))
        if not v:
            log.info("[Failure] request verification failed.")
            log.info("Valid client: %s", valid_client)
            log.info("Valid signature: %s", valid_signature)

        if valid_client and not valid_signature:                                                          # TOOPHER
            norm_params = signature.normalize_parameters(request.params)                                  # TOOPHER
            uri = signature.normalize_base_string_uri(request.uri)                                        # TOOPHER
            base_signing_string = signature.construct_base_string(request.http_method, uri, norm_params)  # TOOPHER
            self.validation_error_message = \
                'Invalid signature. Expected signature base string: {0}'.format(base_signing_string)      # TOOPHER
        return v, request

    def _check_signature(self, request):
        if request.signature_method == SIGNATURE_RSA:  # pragma: nocover
            key_str = self.request_validator.get_rsa_key(request.client_key, request)
            key_str = preprocess_rsa_key(key_str)
            return signature.verify_rsa_sha1(request, key_str)

        return super(TweakedSignatureOnlyEndpoint, self)._check_signature(request)



#
# Copyright (c) 2019, salesforce.com, inc.
# All rights reserved.
# SPDX-License-Identifier: BSD-3-Clause
# For full license text, see the LICENSE file in the repo root or https://opensource.org/licenses/BSD-3-Clause
#

import unittest

from django_declarative_apis.authentication.oauthlib import oauth_errors


class OAuthErrorsTestCase(unittest.TestCase):
    def test_oauth_invalid_version_error(self):
        err = oauth_errors.OAuthInvalidVersionError()
        self.assertEqual(err.detail, 'OAuth version (oauth_version parameter) should be 1.0')
        self.assertEqual(
            err.auth_header,
            'OAuth realm="API",oauth_problem=version_rejected&oauth_acceptable_versions=1.0-1.0'
        )

    def test_build_error(self):
        self.assertIsInstance(
            oauth_errors.build_error('...Timestamp given is invalid...'),
            oauth_errors.OAuthTimestampError
        )

        self.assertIsInstance(
            oauth_errors.build_error('...parameter_absent...foo:bar'),
            oauth_errors.OAuthMissingParameterError
        )

        self.assertIsInstance(
            oauth_errors.build_error('...Invalid signature...'),
            oauth_errors.OAuthInvalidSignatureError
        )

        self.assertIsInstance(
            oauth_errors.build_error('...Invalid OAuth version...'),
            oauth_errors.OAuthInvalidVersionError
        )

        # an unknown error happened
        generic_error = oauth_errors.build_error('something bad happened')
        self.assertIsInstance(generic_error, oauth_errors.OAuthError)
        self.assertEqual(generic_error.detail, 'something bad happened')
        self.assertEqual(generic_error.auth_header, 'OAuth realm="API",oauth_problem=something bad happened')

        # a known error
        for key, msg in oauth_errors.error_to_human_readable_message.items():
            generic_error = oauth_errors.build_error(key)
            self.assertIsInstance(generic_error, oauth_errors.OAuthError)
            self.assertEqual(generic_error.detail, msg)
            self.assertEqual(generic_error.auth_header, f'OAuth realm="API",oauth_problem={key}')



#
# Copyright (c) 2019, salesforce.com, inc.
# All rights reserved.
# SPDX-License-Identifier: BSD-3-Clause
# For full license text, see the LICENSE file in the repo root or https://opensource.org/licenses/BSD-3-Clause
#

from django.conf.urls import url

from . import views

from django_declarative_apis.adapters import resource_adapter

UUID4_REGEX = r'[0-9a-fA-F]{8}-([0-9a-fA-F]{4}-){3}[0-9a-fA-F]{12}'

urlpatterns = [
    url(r'^simple', resource_adapter(get=views.SimpleEndpointDefinition)),
]



#
# Copyright (c) 2019, salesforce.com, inc.
# All rights reserved.
# SPDX-License-Identifier: BSD-3-Clause
# For full license text, see the LICENSE file in the repo root or https://opensource.org/licenses/BSD-3-Clause
#

import time

from django_declarative_apis.authentication import AuthenticationFailure


class OAuthError(AuthenticationFailure):
    detail = None
    auth_header = None


class OAuthTimestampError(OAuthError):
    def __init__(self):
        self.detail = 'There was a problem with your timestamp. Please check your current system time.' \
                      'Server time is {0}.'.format(time.time())
        self.auth_header = 'OAuth realm="API",oauth_problem=timestamp_refused&oauth_acceptable_timestamps={0}-{1}' \
            .format(int(time.time()) - 300, int(time.time()))


class OAuthMissingParameterError(OAuthError):
    def __init__(self, detail):
        self.detail = 'Parameters missing: {0}'.format(detail.split(':')[1])
        self.auth_header = 'OAuth realm="API",oauth_problem=parameter_absent&oauth_parameters_absent={0}'.format(
            detail.split(':')[1])


class OAuthInvalidSignatureError(OAuthError):
    def __init__(self, detail):
        self.detail = '{0}.'.format(detail)
        self.auth_header = 'OAuth realm="API",oauth_problem=signature_invalid'


class OAuthInvalidVersionError(OAuthError):
    def __init__(self):
        self.detail = 'OAuth version (oauth_version parameter) should be 1.0'
        self.auth_header = 'OAuth realm="API",oauth_problem=version_rejected&oauth_acceptable_versions=1.0-1.0'


"""
Map internal OAuth errors to those specified in the OAuth Problem Reporting proposal
http://wiki.oauth.net/w/page/12238543/ProblemReporting
"""

error_to_human_readable_message = {
    'nonce_used': 'Nonce was already used',
    'signature_method_rejected': 'Signature method must be "HMAC-SHA1" or "RSA-SHA1',
    'consumer_key_unknown': 'Invalid consumer key or secret',
}


def build_error(error_message):
    if ('Timestamp given is invalid' in error_message) or ('Invalid timestamp' in error_message):
        return OAuthTimestampError()
    if 'parameter_absent' in error_message:
        return OAuthMissingParameterError(error_message)
    if 'Invalid signature' in error_message:
        return OAuthInvalidSignatureError(error_message)
    if 'Invalid OAuth version' in error_message:
        return OAuthInvalidVersionError()

    detail = error_to_human_readable_message.get(error_message, error_message)
    auth_header = 'OAuth realm="API",oauth_problem={0}'.format(error_message)

    return OAuthError(detail=detail, auth_header=auth_header)



#
# Copyright (c) 2019, salesforce.com, inc.
# All rights reserved.
# SPDX-License-Identifier: BSD-3-Clause
# For full license text, see the LICENSE file in the repo root or https://opensource.org/licenses/BSD-3-Clause
#

import abc
import typing


class AuthenticatorHint(typing.NamedTuple):
    """ Tuple to provide hints for authentication implementations

    header_hint: A string used to match the `Authentication: ` header. 
    """
    header: str


class Authenticator(metaclass=abc.ABCMeta):
    @abc.abstractmethod
    def is_authenticated(self, request):
        raise NotImplementedError

    @abc.abstractmethod
    def challenge(self, error):
        """ Results in the challenge response sent to the user

        This should result in a django.http.HttpResponse that should include information through the
        WWW-Authenticate header around expectations.
        """
        raise NotImplementedError


class AuthenticationResult(object):
    def __init__(self, detail=None, auth_header=None):
        self.detail = detail
        self.auth_header = auth_header


class NoAuthentication(AuthenticationResult, Authenticator):
    """
    Authentication handler that always returns
    True, so no authentication is needed, nor
    initiated (`challenge` is missing.)
    """

    def is_authenticated(self, request):
        return True


class AuthenticationSuccess(AuthenticationResult):
    def __bool__(self):
        return True


class AuthenticationFailure(AuthenticationResult):
    def __bool__(self):
        return False


def validate_authentication_config(config):
    """ Validate the computed configuration of authentication handlers

    The schema for the config is:

    {
        <AuthenticatorHint>: [<Authenticator>, <Authenticator>...],
        <AuthenticatorHint>: [<Authenticator>, <Authenticator>...],
    }

    AuthenticatorHints allow us to match Authorization headers for quick handler lookup. For example,
    if we want to use OAuth 1.0a, we could use an AuthenticatorHint.header value of 'OAuth ' as a key
    and a value of [django_declarative_apis.authentication.oauthilib.oauth1.TwoLeggedOauth1()]. This will
    ensure that any time an `Authorization: OAuth ...` header is seen, the appropriate authenticator is used.

    If there are more complexities to the authenticator (i.e. OAuth 1.0a allows for the transport of credentials
    through header, request body or query parameters, catch-alls are allowed by using a key of `None`:

    {
        None: [<Authenticator>],
    }

    The catch-all authenticators are always executed after matched authenticators.

    Note: This may need to get smarter in the future but was kept simple intentionally as it's executed on
          every request.
    """
    assert isinstance(config, typing.Mapping)
    for hint, authenticators in config.items():
        if not isinstance(hint, (AuthenticatorHint, type(None))):
            raise TypeError(
                'Authenticator hint must be an instance of authentication.AuthenticatorHint or None'
            )

        assert isinstance(authenticators, (list, tuple))
        for authenticator in authenticators:
            if not isinstance(authenticator, Authenticator):
                raise TypeError(
                    'Authenticator must be an instance of authentication.Authenticator'
                )



#!/usr/bin/env python
#
# Copyright (c) 2019, salesforce.com, inc.
# All rights reserved.
# SPDX-License-Identifier: BSD-3-Clause
# For full license text, see the LICENSE file in the repo root or https://opensource.org/licenses/BSD-3-Clause
#

import setuptools
from os import path

with open(path.join(path.dirname(__file__), 'requirements.txt')) as requirements_file:
    requirements = requirements_file.read().split()

setuptools.setup(
    name='django-declarative-apis',
    version='0.19.2',
    author='Drew Shafer',
    url='https://salesforce.com',
    description='Simple, readable, declarative APIs for Django',
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Environment :: Web Environment',
        'Framework :: Django',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Topic :: Internet :: WWW/HTTP',
        ],
    packages=setuptools.find_packages(),
    test_suite='tests',
    install_requires=requirements
)



#
# Copyright (c) 2019, salesforce.com, inc.
# All rights reserved.
# SPDX-License-Identifier: BSD-3-Clause
# For full license text, see the LICENSE file in the repo root or https://opensource.org/licenses/BSD-3-Clause
#

from django.http import HttpResponse

from django_declarative_apis.machinery import EndpointDefinition, field


class SimpleEndpointDefinition(EndpointDefinition):
    def is_authorized(self):
        return True

    int_type_field = field(type=int)

    @property
    def resource(self):
        return {}



#
# Copyright (c) 2019, salesforce.com, inc.
# All rights reserved.
# SPDX-License-Identifier: BSD-3-Clause
# For full license text, see the LICENSE file in the repo root or https://opensource.org/licenses/BSD-3-Clause
#

import unittest

import django.test
from django_declarative_apis.authentication.oauthlib import endpoint, request_validator
from django.core.cache import cache
from django_declarative_apis import models
from oauthlib.oauth1 import rfc5849
from tests import testutils


# TODO: create tests for android and ios rsa signing


def _oauth_qs(request):
    return '&'.join(['='.join((k, v)) for k, v in request.GET.items()])

class TweakedSignatureOnlyEndpointTestCase(django.test.TestCase):
    def _request(self, url, method='GET', rsa_key=None):
        request = getattr(self.request, method.lower())(url)
        request.META['consumer'] = self.consumer
        if rsa_key is not None:
            request.META['rsa_key'] = rsa_key
        testutils.OAuthClientHandler._build_request(request)

        validator = request_validator.DjangoRequestValidator(request)
        endpoint_instance = endpoint.TweakedSignatureOnlyEndpoint(validator)

        return request, validator, endpoint_instance

    def setUp(self):
        self.request = django.test.RequestFactory()
        self.consumer = models.OauthConsumer.objects.create(name='test')
        self.consumer.save()

    def test_validate_request_missing_parameters_error(self):
        request, validator, endpoint_instance = self._request('/test')
        result = endpoint_instance.validate_request('/test')
        
        self.assertEqual(result, (False, None))

    def test_validate_request_invalid_transport_security_error(self):
        with self.settings(REQUIRE_HTTPS_FOR_OAUTH=True):
            request, validator, endpoint_instance = self._request('/test')
            result, _ = endpoint_instance.validate_request(f"http://test?{_oauth_qs(request)}")

        self.assertEqual(endpoint_instance.validation_error_message, 'Only HTTPS connections are permitted.')
        self.assertFalse(result)

    def test_validate_request_invalid_parameters_error(self):
        request, validator, endpoint_instance = self._request('/test')

        params = request.GET.copy()
        del params['oauth_signature_method']
        request.GET = params

        result, _ = endpoint_instance.validate_request(f"http://test?{_oauth_qs(request)}")
        
        self.assertEqual(endpoint_instance.validation_error_message, 'Missing mandatory OAuth parameters.')
        self.assertFalse(result)

    def test_validate_request_invalid_nonce_error(self):
        request, validator, endpoint_instance = self._request('/test')

        # nonce has already been used
        cache.set(f"{request.META['consumer'].key}:::{request.GET['oauth_nonce']}", True)
        result, _ = endpoint_instance.validate_request(f'http://test?{_oauth_qs(request)}')

        self.assertEqual(validator.validation_error_message, 'nonce_used')
        self.assertFalse(result)

    def test_validate_request_invalid_client_error(self):
        request, validator, endpoint_instance = self._request('/test')

        params = request.GET.copy()
        params['oauth_consumer_key'] = 'bad'
        request.GET = params

        result, _ = endpoint_instance.validate_request(f'http://test?{_oauth_qs(request)}')
        self.assertFalse(result)

    def test_validate_request_invalid_signature_error(self):
        request, validator, endpoint_instance = self._request('/test')

        params = request.GET.copy()
        params['oauth_signature'] = 'bad'
        request.GET = params

        result, _ = endpoint_instance.validate_request(f'http://test?{_oauth_qs(request)}')
        self.assertTrue(endpoint_instance.validation_error_message.startswith('Invalid signature. Expected signature base string:'))
        self.assertFalse(result)



#
# Copyright (c) 2019, salesforce.com, inc.
# All rights reserved.
# SPDX-License-Identifier: BSD-3-Clause
# For full license text, see the LICENSE file in the repo root or https://opensource.org/licenses/BSD-3-Clause
#

from django.db import models


class TestModel(models.Model):
    int_field = models.IntegerField()

    EXPANDABLE_DICT_RETURN = {'key': 'value'}
    EXPANDABLE_STRING_RETURN = 'string'

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def mutate_action(self):
        self.int_field += 1
        self.save()

    @property
    def expandable_dict(self):
        return TestModel.EXPANDABLE_DICT_RETURN

    @property
    def expandable_string(self):
        return TestModel.EXPANDABLE_STRING_RETURN


class ChildModel(models.Model):
    id = models.IntegerField(primary_key=True)
    name = models.CharField(max_length=100)
    test = models.ForeignKey(TestModel, on_delete=models.CASCADE)
    parent = models.ForeignKey('ParentModel', related_name='children', on_delete=models.CASCADE)


class ParentModel(models.Model):
    nonstandard_id = models.IntegerField(primary_key=True)
    name = models.CharField(max_length=100)
    favorite = models.ForeignKey(ChildModel, null=True, on_delete=models.CASCADE)
    root = models.ForeignKey('RootNode', related_name='parents', null=True, on_delete=models.CASCADE)


class RootNode(models.Model):
    id = models.IntegerField(primary_key=True)
    parent_field = models.ForeignKey(ParentModel, on_delete=models.CASCADE)
